/*
 * yasapi_engine_client.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static int EngineClientPreOpen(EngineClient *pClient, int bIncBroken,
    int bIncTracks);
static EEngineServer EngineClientServerState(EngineClient *pClient);

///////////////////////////////////////////////////////////////////////////////
static int EngineClientReady(EngineClient *pClient)
{
  int bReady=0;
  Ring *pRing=&pClient->ring;
  WORD uFramesPlaybackBegin;
  WORD uFramesWritten;

  if (FsmState(pClient)<eEngineClientStateRing)
    goto exit;

  uFramesPlaybackBegin=pClient->param.uFramesPlaybackBegin;
  uFramesWritten=pRing->frames._uWritten;
  bReady=uFramesWritten<uFramesPlaybackBegin?0:1;
exit:
  return bReady;
}

///////////////////////////////////////////////////////////////////////////////
static void EngineClientUpdateWritten(EngineClient *pClient)
{
  Ring *pRing=&pClient->ring;
  double xWritten=pRing->frames._uWritten;
  double xSize=pRing->frames._uWritten+pRing->frames._uAvailable;

  pClient->monitor.xTailCur=xWritten/xSize;

  // xTailCurMin:
  if (YASAPI_DOUBLE_INVALID(pClient->monitor.xTailMin))
    pClient->monitor.xTailMin=pClient->monitor.xTailCur;
  else if (pClient->monitor.xTailCur<pClient->monitor.xTailMin)
    pClient->monitor.xTailMin=pClient->monitor.xTailCur;

  // xTailCurMax:
  if (YASAPI_DOUBLE_INVALID(pClient->monitor.xTailMax))
    pClient->monitor.xTailMax=pClient->monitor.xTailCur;
  else if (pClient->monitor.xTailMax<pClient->monitor.xTailCur)
    pClient->monitor.xTailMax=pClient->monitor.xTailCur;
}

///////////////////////////////////////////////////////////////////////////////
int EngineClientCreate(EngineClient *pClient, Player *pPlayer)
{
  /////////////////////////////////////////////////////////////////////////////
  memset(pClient,0,sizeof *pClient);
  FsmCreate(pClient,gcaIEngineClientState ,&pPlayer->gtk,
      &pPlayer->gtk.config.config.monitor.state.client);

  /////////////////////////////////////////////////////////////////////////////
  pClient->pPlayer=pPlayer;
  EngineMonitorReset(&pClient->monitor);

  return 0;
}

void EngineClientDestroy(EngineClient *pClient)
{
  YASAPI_PLAYER_LOCK(pClient->pPlayer); // {
  FsmDestroy(pClient);
  YASAPI_PLAYER_UNLOCK(pClient->pPlayer); // }
}

///////////////////////////////////////////////////////////////////////////////
int EngineClientRequestServerState(EngineClient *pClient,
    EEngineServer eRequest, const char *file, int line, const char *func)
{
  int code=-1;
  EngineServer *pServer=pClient->pServer;
  Player *pPlayer=pClient->pPlayer;
  int nLockRecursion;
  EEngineServer eEngineServer;
  DWORD dwWait;

  YASAPI_PLAYER_LOCK(pPlayer); // {
  pServer->thread.eStateRequest=eRequest;
  DVWRITELNEX(file,line,func,"### %s {",FsmStateIdEx(pServer,eRequest));
  SetEvent(pServer->thread.hEvent);

  for (;;) {
    eEngineServer=FsmState(pServer);

    if (eEngineServer==eRequest)
      break;
    else if (eEngineServer==eEngineServerStateError)
      goto exit;

    ResetEvent(pClient->thread.hEvent);
    YASAPI_PLAYER_UNLOCK_DEEP(pPlayer,&nLockRecursion); // {
    DWRITELNEX(file,line,func,"{");
    dwWait=WaitForSingleObject(pClient->thread.hEvent,10000ul);
    DWRITELNEX(file,line,func,"}");
    YASAPI_PLAYER_LOCK_DEEP(pPlayer,nLockRecursion); // }

    switch (dwWait) {
    case WAIT_ABANDONED:
      DVMESSAGEEX(file,line,func,"WAIT_ABANDONED %s %s",
          FsmStateIdEx(pServer,eRequest),FsmStateId(pServer));
      break;
    case WAIT_TIMEOUT:
      DVMESSAGEEX(file,line,func,"WAIT_TIMEOUT %s %s",
          FsmStateIdEx(pServer,eRequest),FsmStateId(pServer));
      break;
    case WAIT_FAILED:
      DVMESSAGEEX(file,line,func,"WAIT_FAILED %s %s: %lu",
          FsmStateIdEx(pServer,eRequest),FsmStateId(pServer),GetLastError());
      break;
    default:
      break;
    }
  }

  DVWRITELNEX(file,line,func,"### %s }",FsmStateIdEx(pServer,eRequest));
  code=0;
exit:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
  return code;
}


///////////////////////////////////////////////////////////////////////////////
#define YASAPI_START_DEBUG
static int EngineClientDrain(EngineClient *pClient)
{
  Player *pPlayer=pClient->pPlayer;
  int nLockRecursion;
  DWORD dwWait;

  while (eEngineServerStateKill<FsmState(pClient->pServer)) {
    ResetEvent(pClient->thread.hEvent);
    YASAPI_PLAYER_UNLOCK_DEEP(pPlayer,&nLockRecursion); // {
    dwWait=WaitForSingleObject(pClient->thread.hEvent,10000ul);
    YASAPI_PLAYER_LOCK_DEEP(pPlayer,nLockRecursion); // }

    switch (dwWait) {
    case WAIT_ABANDONED:
      DVMESSAGE("WAIT_ABANDONED %s",FsmStateId(pClient->pServer));
      goto exit;
    case WAIT_TIMEOUT:
      DVMESSAGE("WAIT_TIMEOUT %s",FsmStateId(pClient->pServer));
      goto exit;
    case WAIT_FAILED:
      DVMESSAGE("WAIT_FAILED %s: %lu",FsmStateId(pClient->pServer),
          GetLastError());
      goto exit;
    default:
      break;
    }
  }

  return 0;
exit:
  return -1;
}

int EngineClientOpen(EngineClient *pClient, int samplerate, int numchannels,
    int bitspersamp, int *pnMaxLatency, int *pnFramesDevice)
{
  int code=-1;
  Player *pPlayer=pClient->pPlayer;
  GtkClient *pGtk=&pPlayer->gtk;
  int bFirst=1;
  int bReady=0;
  int bIncGaplessBroken=1;
  EngineOptions options;
  int bOptionsChanged;
  EngineWfx wfx;
  int bTargetChanged;
#if 0 // [
  int nLockRecursion;
  DWORD dwWait;
#endif // ]

  ///////////////////////////////////////////////////////////////////////////
  YASAPI_PLAYER_LOCK(pPlayer); // {
DVWRITELN("=== %p >>>=====================================",GetCurrentThread());
  bFirst=FsmState(pClient)<eEngineClientStateBase;
  bReady=EngineClientReady(pClient);

  /////////////////////////////////////////////////////////////////////////////
  if (bFirst&&YASAPI_FSM_UP(pClient,eEngineClientStateBase)<0) {
    DMESSAGE("creating producer base");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (EngineOptionsCreate(&options,pClient)<0) {
    DMESSAGE("creating options");
    goto unlock;
  }

  bOptionsChanged=memcmp(&options,&pClient->options, sizeof options);

  /////////////////////////////////////////////////////////////////////////////
  if (EngineWfxCreate(&wfx,&options,samplerate,numchannels,bitspersamp)<0) {
    DMESSAGE("creating WAVEFORMATEXTENSIBLEs");
    goto unlock;
  }

  bTargetChanged=memcmp(&wfx.target.Format,&pClient->wfx.target.Format,
      sizeof wfx.target.Format);

  /////////////////////////////////////////////////////////////////////////////
  if (!options.common.bGapless) {
#if defined (YASAPI_START_DEBUG) // [
    DWRITELN(">>> bGapless <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
#endif // ]
    goto start;
  }

  if (bOptionsChanged) {
#if defined (YASAPI_START_DEBUG) // [
    DWRITELN(">>> bOptionsChanged <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
#endif // ]
    goto start;
  }

  if (bTargetChanged) {
#if defined (YASAPI_START_DEBUG) // [
    DWRITELN(">>> bTargetChanged <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
#endif // ]
    goto start;
  }

  if (FsmState(pClient->pServer)==eEngineServerStateBase) {
#if defined (YASAPI_START_DEBUG) // [
    DWRITELN(">>> eEngineServerStateBase <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
#endif // ]
    goto start;
  }

  if (FsmState(pClient->pServer)==eEngineServerStateKill) {
#if defined (YASAPI_START_DEBUG) // [
    DWRITELN(">>> eEngineServerStateKill <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
#endif // ]
    goto start;
  }

  if (!bReady) {
#if defined (YASAPI_START_DEBUG) // [
    DWRITELN(">>> bReady <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
#endif // ]
    goto start;
  }

#if defined (YASAPI_START_DEBUG) // [
  DWRITELN(">>> Continue Gaplessly <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
#endif // ]
  // Continue Gaplessly ///////////////////////////////////////////////////////
  ++pClient->pPlayer->stats.uNumTracks;
  ++pClient->stats.uNumTracks;
  EngineClientUpdateWritten(pClient);
  pClient->param.time.lpVtbl->Inc(pClient->pServer);
  YASAPI_FSM_DOWN(pClient->pServer,eEngineServerStateOpen);

  if (RingResetWrite(&pClient->ring,0)<0) {
    DMESSAGE("resetting ring write");
    goto exit;
  }

  memcpy(&pClient->wfx.source,&wfx.source,sizeof pClient->wfx.source);
  goto success;
start:
DVWRITELN("%s ============================",FsmStateId(pClient));
  // Start ////////////////////////////////////////////////////////////////////
  bIncGaplessBroken=bIncGaplessBroken
      &&pClient->options.common.bGapless;
  bIncGaplessBroken=bIncGaplessBroken
      &&!bTargetChanged;
  bIncGaplessBroken=bIncGaplessBroken
      &&bOptionsChanged;

  if (!bFirst) {
    bIncGaplessBroken=bIncGaplessBroken
        ||FsmState(pClient->pServer)<=eEngineServerStateKill;
    bIncGaplessBroken=bIncGaplessBroken
        ||bReady;
  }

#if 0 // [
  while (eEngineServerStateKill<FsmState(pClient->pServer)) {
    ResetEvent(pClient->thread.hEvent);
    YASAPI_PLAYER_UNLOCK_DEEP(pPlayer,&nLockRecursion); // {
    dwWait=WaitForSingleObject(pClient->thread.hEvent,10000ul);
    YASAPI_PLAYER_LOCK_DEEP(pPlayer,nLockRecursion); // }

    switch (dwWait) {
    case WAIT_ABANDONED:
      DVMESSAGE("WAIT_ABANDONED %s",FsmStateId(pClient->pServer));
      goto exit;
    case WAIT_TIMEOUT:
      DVMESSAGE("WAIT_TIMEOUT %s",FsmStateId(pClient->pServer));
      goto exit;
    case WAIT_FAILED:
      DVMESSAGE("WAIT_FAILED %s: %lu",FsmStateId(pClient->pServer),
          GetLastError());
      goto exit;
    default:
      break;
    }
  }
#else // ] [
DMARKLN();
  if (EngineClientDrain(pClient)<0)
    goto exit;
#endif // ]

  YASAPI_FSM_DOWN(pClient,eEngineClientStateBase);
  memcpy(&pClient->options,&options,sizeof pClient->options);
  memcpy(&pClient->wfx,&wfx,sizeof pClient->wfx);

  ///////////////////////////////////////////////////////////////////////////
  if (EngineClientPreOpen(pClient,bIncGaplessBroken,1)<0) {
    DMESSAGE("creating gapless sequence");
    goto exit;
  }
success:
  pClient->winamp.hThread=GetCurrentThread();
  pClient->winamp.bAltered=0;
  GtkClientUpdateDialog(pGtk,&pGtk->config);

  pClient->param.nMaxLatency=2.0*pClient->options.device.xMaxLatencyAssumed+0.5;

  if (pnMaxLatency)
    *pnMaxLatency=pClient->param.nMaxLatency;

  if (pnFramesDevice)
    *pnFramesDevice=pClient->param.uNumFramesDevice;

  code=0;
unlock:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
exit:
  return code;
}

static void _EngineClientCloseInternal(EngineClient *pClient)
{
  if (!pClient->param.bActive) {
DMARKLN();
    EngineClientDrain(pClient);
  }
  else {
DVWRITELN("2 %s {",FsmStateIdEx(pClient,eEngineClientStateBase));
  YASAPI_REQUEST_STATE(pClient,eEngineServerStateBase);
DVWRITELN("2 %s }",FsmStateIdEx(pClient,eEngineClientStateBase));
  }

  memset(&pClient->wfx.source,0,sizeof pClient->wfx.source);
  memset(&pClient->wfx.target,0,sizeof pClient->wfx.target);
  memset(&pClient->param,0,sizeof pClient->param);
}

static void _EngineClientClose(EngineClient *pClient)
{
  int bKill=0;
  EEngineClient eEngineClient=FsmState(pClient);
  EEngineServer eEngineServer=FsmState(pClient->pServer);

  DVWRITELN("=== target: %u/%u/%u ===",
      pClient->wfx.target.Format.nSamplesPerSec,
      pClient->wfx.target.Format.nChannels,
      pClient->wfx.target.Format.wBitsPerSample);
  DVWRITELN("eEngineServer: %s (%d, %d, %d)",pStateId,eEngineServerStateOpen,eEngineServer,eEngineServerStateOpen<=eEngineServer);

  if (eEngineServerStateKill==eEngineServer||eEngineServerStateOpen<=eEngineServer) {
DVWRITELN("bKill: %d",bKill);
    bKill=bKill||!pClient->param.bActive;
DVWRITELN("bKill: %d (!pClient->param.bActive)",bKill);
    bKill=bKill||(eEngineServerStateOpen<=eEngineServer&&eEngineServer<eEngineServerStateDraining);
DVWRITELN("bKill: %d (eEngineServerStateOpen<=%s&&%s<eEngineServerStateDraining)",bKill,pStateId,pStateId);
    bKill=bKill||(eEngineServerStateKill==eEngineServer);
DVWRITELN("bKill: %d (eEngineServerStateKill==%s)",bKill,pStateId);
    bKill=bKill||!pClient->options.common.bGapless;
DVWRITELN("bKill: %d (!pClient->options.common.bGapless)",bKill);

    if (bKill) {
DWRITELN("=== KILLING ==================================================");
DWRITELN(FsmStateId(pClient));
      _EngineClientCloseInternal(pClient);
DWRITELN(FsmStateId(pClient));

      if (eEngineClient<eEngineClientStateClose) {
DWRITELN(FsmStateId(pClient));
        YASAPI_FSM_UP(pClient,eEngineClientStateClose);
DWRITELN(FsmStateId(pClient));
      }
      else if (eEngineClientStateClose<eEngineClient) {
DWRITELN(FsmStateId(pClient));
        YASAPI_FSM_DOWN(pClient,eEngineClientStateClose);
DWRITELN(FsmStateId(pClient));
      }
DWRITELN(FsmStateId(pClient));
    }
  }
}

void EngineClientClose(EngineClient *pClient)
{
  Player *pPlayer=pClient->pPlayer;

  if (FsmState(pClient)<eEngineClientStateBase) {
    DMESSAGE("not initialized");
    goto enoinit;
  }

  YASAPI_PLAYER_LOCK(pPlayer); // {
DVWRITELN("=== %p <<<=====================================",GetCurrentThread());

  if (eEngineClientStateClose<FsmState(pClient))
    _EngineClientClose(pClient);

  YASAPI_PLAYER_UNLOCK(pPlayer); // }
enoinit: ;
}

///////////////////////////////////////////////////////////////////////////////
static int _EngineClientAvailable(EngineClient *pClient)
{
  Ring *pRing=&pClient->ring;
  int nAvailable=0;

  if (FsmState(pClient)<eEngineClientStateRing)
    nAvailable=INT_MAX;
  else {
    switch (FsmState(pClient->pServer)) {
    case eEngineServerStateKill:
      nAvailable=INT_MAX;
      break;
    case eEngineServerStatePauseConnected:
    case eEngineServerStatePauseDisconnected:
      break;
    default:
      nAvailable=RingSourceSizeAvailable(pRing);
      break;
    }
  }

  return nAvailable;
}

int EngineClientAvailable(EngineClient *pClient)
{
  int nAvailable=0;
  Player *pPlayer=pClient->pPlayer;

  if (FsmState(pClient)<eEngineClientStateBase)
    goto ebase;

  YASAPI_PLAYER_LOCK(pPlayer); // {
  nAvailable=_EngineClientAvailable(pClient);
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
ebase:
  return nAvailable;
}

static int _EngineClientWrite(EngineClient *pClient, const BYTE *buf, DWORD dwChunkSizeSource)
{
  int code=-1;
  EEngineClient eEngineClient=FsmState(pClient);
  EEngineServer eEngineServer=FsmState(pClient->pServer);
  Player *pPlayer=pClient->pPlayer;
  GtkClient *pGtk=&pPlayer->gtk;
  Ring *pRing=&pClient->ring;
  WORD uFramesPlaybackBegin;
  WORD uFramesWritten;
  WORD uFramesMax;
  DWORD dwSourceSize;
  UINT32 uMsLatency;

  if (eEngineClient<eEngineClientStateBase) {
    DMESSAGE("producer not initialized");
    goto ebase;
  }
  else if (eEngineClient<eEngineClientStatePreOpen) {
    if (EngineClientPreOpen(pClient,1,0)<0) {
      DMESSAGE("creating gapless sequence");
      goto exit;
    }
  }
  else if (eEngineServer==eEngineServerStateKill)
    goto exit;

  switch (eEngineClient) {
  default:
    if (eEngineClient<eEngineClientStateRing) {
      pClient->param.dwChunkSizeSource=dwChunkSizeSource;
      eEngineClient=YASAPI_FSM_UP(pClient,eEngineClientStateRing);

      if (eEngineClient<0) {
        DMESSAGE("creating ring");
        goto exit;
      }
    }
    else if (pClient->param.dwChunkSizeSource!=dwChunkSizeSource) {
      // This is special to "in_mp3" because it has not a constant block size.
      dwSourceSize=RingSourceSizeAvailable(pRing)+dwChunkSizeSource;

      if (RingSourceSize(pRing)<dwSourceSize) {
        if (EngineClientCreateRing(pClient,RingReallocFrames,dwChunkSizeSource)<0) {
          DMESSAGE("reallocating ring");
          goto exit;
        }
      }
    }

    break;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (RingWrite(pRing,buf,dwChunkSizeSource)<0) {
    DMESSAGE("writing to ring");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  uFramesPlaybackBegin=pClient->param.uFramesPlaybackBegin;
  uFramesWritten=pRing->frames._uWritten;
  uMsLatency=YASAPI_FRAMES2MS(uFramesWritten,&pClient->wfx.source)+0.5;

  if (pClient->stats.uNumFramesMax<uFramesWritten) {
    pClient->stats.uNumFramesMax=uFramesWritten;
    GtkClientUpdateWidget(pGtk,
        &pGtk->config.config.monitor.engine.num_frames_latency);
  }

  /////////////////////////////////////////////////////////////////////////////
  if (UINT32_MAX==pPlayer->stats.uMsLatencyMax
      ||pPlayer->stats.uMsLatencyMax<uMsLatency) {
    pPlayer->stats.uMsLatencyMax=uMsLatency;
    GtkClientUpdateWidget(pGtk,
        &pGtk->config.config.monitor.player.ms_latency);
  }

  /////////////////////////////////////////////////////////////////////////////
  if (eEngineClientStateOpen<=eEngineClient&&eEngineServerStateOpen<=eEngineServer) {
    goto success;
  }

  uFramesPlaybackBegin=pClient->param.uFramesPlaybackBegin;
  uFramesWritten=pRing->frames._uWritten;
  if (uFramesWritten<uFramesPlaybackBegin) {
    ///////////////////////////////////////////////////////////////////////////
    DVWRITELN("*** underflow: %u %u (%u %lu)",uFramesWritten,
        _EngineClientAvailable(pClient),uFramesPlaybackBegin,
        dwChunkSizeSource);
    //uFramesMax=pRing->lpVtbl->FramesMaximum(pRing);
    uFramesMax=RingFramesMaximum(pRing);
    pClient->monitor.xWritten=(double)uFramesWritten/uFramesMax;
    GtkClientUpdateWidget(pGtk,&pGtk->config.config.monitor.written);
    goto success;
  } 
  else if (eEngineServer<eEngineServerStateOpen) {
    DVWRITELN("*** starting:  %u %u (%u %lu)",uFramesWritten,
        _EngineClientAvailable(pClient),uFramesPlaybackBegin,
        dwChunkSizeSource);

    if (YASAPI_FSM_UP(pClient->pServer,eEngineServerStateOpen)<0) {
      DMESSAGE("start rendering");
      goto exit;
    }
  }
success:
  code=0;
exit:
ebase:
  return code;
}

int EngineClientWrite(EngineClient *pClient, BYTE *buf, DWORD dwChunkSize)
{
  Player *pPlayer=pClient->pPlayer;
  int code;

  YASAPI_PLAYER_LOCK(pPlayer); // {
  code=_EngineClientWrite(pClient,buf,dwChunkSize);
  YASAPI_PLAYER_UNLOCK(pPlayer); // }

  return code;
}

///////////////////////////////////////////////////////////////////////////////
static int _EngineClientIsPlaying(EngineClient *pClient)
{
  int bIsPlaying=0;
  EngineServer *pServer=pClient->pServer;

  switch (FsmState(pServer)) {
  case eEngineServerStatePauseDisconnected:
  case eEngineServerStatePauseConnected:
  case eEngineServerStateOpen:
    if (YASAPI_FSM_UP(pServer,eEngineServerStateDraining)<0) {
      DMESSAGE("start draining");
      goto exit;
    }

    // Intentional fall-through.
  case eEngineServerStateDraining:
  case eEngineServerStateEot:
    bIsPlaying=pClient->options.common.bGapless?0:
        0ul<RingTargetSize(&pClient->ring)?1:
        0;
    break;
  default:
    break;
  }
exit:
  return bIsPlaying;
}

int EngineClientIsPlaying(EngineClient *pClient)
{
  Player *pPlayer=pClient->pPlayer;
  int bIsPlaying=0;

  if (FsmState(pClient)<eEngineClientStateBase)
    goto eopen;

  YASAPI_PLAYER_LOCK(pPlayer); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    goto exit;

  bIsPlaying=_EngineClientIsPlaying(pClient);
exit:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
eopen:
  return bIsPlaying;
}

///////////////////////////////////////////////////////////////////////////////
int EngineClientFlush(EngineClient *pClient, int ms)
{
  int code=-1;
  Player *pPlayer=pClient->pPlayer;
  GtkClient *pGtk=&pPlayer->gtk;
  ConfigMonitor *pMonitor=&pGtk->config.config.monitor;
  EngineServer *pServer=pClient->pServer;
  IAudioClient *pAudioClient;
  HRESULT hr;

  if (FsmState(pClient)<eEngineClientStateBase) {
    DMESSAGE("something went wrong!");
    goto eopen;
  }

  YASAPI_PLAYER_LOCK(pPlayer); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    goto exit;

  if (FsmState(pClient->pServer)<eEngineServerStateOpen)
    goto success;

  YASAPI_FSM_DOWN(pClient->pServer,eEngineServerStateFlush);
  pAudioClient=pServer->pAudioClient;
  hr=pAudioClient->lpVtbl->Reset(pAudioClient);

  if (FAILED(hr)) {
#define DCLABEL "IAudioClient::Reset" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_INITIALIZED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_STOPPED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_BUFFER_OPERATION_PENDING);
    DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto exit;
  }

  RingReset(&pServer->pClient->ring);
  pClient->monitor.xWritten=0.0;
  GtkClientUpdateWidget(pGtk,&pMonitor->written);
  pClient->param.time.lpVtbl->Set(pClient->pServer,ms);
success:
  code=0;
exit:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
eopen:
  return code;
}

///////////////////////////////////////////////////////////////////////////////
static int _EngineClientPause(EngineClient *pClient, int nPause)
{
  int nLastPause=pClient->param.nLastPause;
  EngineServer *pServer=pClient->pServer;
  EEngineServer eEngineServer=FsmState(pServer);

  if (nLastPause==nPause)
    goto exit;

  if (nPause) {
    switch (eEngineServer) {
    case eEngineServerStateOpen:
      if (pClient->options.common.bDisconnect) {
DWRITELN("3 {");
        YASAPI_FSM_DOWN(pServer,eEngineServerStatePauseDisconnected);
DWRITELN("3 }");
      }
      else {
DWRITELN("4 {");
        YASAPI_FSM_DOWN(pServer,eEngineServerStatePauseConnected);
DWRITELN("4 }");
      }

      pClient->param.nLastPause=nPause;
      break;
    default:
      break;
    }
  }
  else {
    switch (eEngineServer) {
    case eEngineServerStatePauseDisconnected:
    case eEngineServerStatePauseConnected:
DWRITELN("5 {");
        if (YASAPI_FSM_UP(pServer,eEngineServerStateOpen)<0) {
        DMESSAGE("ending pause");
        goto exit;
      }
DWRITELN("5 }");
			if (pClient->pServer)
				SetEvent(pClient->pServer->api.hEvent);

      pClient->param.nLastPause=nPause;
      break;
    default:
      break;
    }
  }
exit:
  return nLastPause;
}

int EngineClientPause(EngineClient *pClient, int nPause)
{
  Player *pPlayer=pClient->pPlayer;
  int nLastPause=0;

  if (FsmState(pClient)<eEngineClientStateBase)
    goto ebase;

  YASAPI_PLAYER_LOCK(pPlayer); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    goto exit;

  nLastPause=_EngineClientPause(pClient,nPause);
exit:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
ebase:
  return nLastPause;
}

///////////////////////////////////////////////////////////////////////////////
// Return time history corrected.
int EngineClientOutputTime(EngineClient *pClient)
{
  int ms=0;
  Player *pPlayer=pClient->pPlayer;
  EngineServer *pServer=pClient->pServer;
  EEngineServer eEngineServer;

  if (pClient->options.common.bLatency)
    goto ebase;
  else if (FsmState(pClient)<eEngineClientStateBase)
    goto ebase;

  YASAPI_PLAYER_LOCK(pPlayer); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    goto exit;

  eEngineServer=FsmState(pServer);

  switch (eEngineServer) {
  case eEngineServerStatePauseDisconnected:
  case eEngineServerStatePauseConnected:
    goto exit;
  default:
    if (eEngineServer<eEngineServerStateOpen)
      goto exit;

    break;
  }

  ms=pClient->param.time.lpVtbl->Get(pServer);
exit:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
ebase:
  return ms;
}

int EngineClientWrittenTime(EngineClient *pClient)
{
  WAVEFORMATEXTENSIBLE *pTarget=&pClient->wfx.target;
  double msWrittenTime=0.0;
  int nLatency=0;
  Player *pPlayer=pClient->pPlayer;
  Ring *pRing=&pClient->ring;
  EngineServer *pServer=pClient->pServer;
  EEngineServer eEngineServer;
  WORD uFramesWritten;
  UINT32 uNumFramesPadding;

  if (FsmState(pClient)<eEngineClientStateBase)
    goto ebase;

  YASAPI_PLAYER_LOCK(pPlayer); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    goto exit;

  eEngineServer=FsmState(pServer);

  switch (eEngineServer) {
  case eEngineServerStatePauseDisconnected:
  case eEngineServerStatePauseConnected:
    goto exit;
  default:
    if (eEngineServer<eEngineServerStateOpen)
      goto exit;

    break;
  }

  uFramesWritten=pRing->frames._uWritten;
  nLatency=YASAPI_FRAMES2MS(uFramesWritten,pTarget);

  if (!pClient->param.bDoubleBuffered) {
    if (EngineServerNumFramesPadding(pServer,&uNumFramesPadding)<0)
      goto exit;

    nLatency+=YASAPI_FRAMES2MS(uNumFramesPadding,&pClient->wfx.target);
  }

  if (pClient->param.nMaxLatency<nLatency)
    nLatency=pClient->param.nMaxLatency;

  if (!pClient->options.common.bLatency)
    msWrittenTime+=pClient->param.time.lpVtbl->Get(pServer);
exit:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
ebase:
  return msWrittenTime+nLatency+0.5;
}

void EngineClientVolume(EngineClient *pClient, Player *pPlayer)
{
  YASAPI_PLAYER_LOCK(pPlayer); // {

  if (FsmState(pClient->pServer)<eEngineServerStateRender)
    gcIEngineServerVolumeNullVtbl.Set(pClient->pServer);
  else
    pClient->param.volume.lpVtbl->Set(pClient->pServer);

  YASAPI_PLAYER_UNLOCK(pPlayer); // }

  return;
}


int EngineClientApiWrite(EngineClient *pClient, int msTimeout, const char *buf,
    DWORD dwChunkSize)
{
  int code=-1;
  Player *pPlayer=pClient->pPlayer;
  HANDLE hEvent=pClient->pServer->api.hEvent;
  int nAvailable;
  int nLockRecursion;
  DWORD dwWait;

  if (FsmState(pClient)<eEngineClientStateBase) {
    DVMESSAGE("unexpected state %s",FsmStateId(pClient));
    goto enoinit;
  }

  YASAPI_PLAYER_LOCK(pPlayer); // {

  for (;;) {
    nAvailable=_EngineClientAvailable(pClient);

    switch (FsmState(pClient->pServer)) {
    case eEngineServerStateKill:
      goto exit;
    default:
      if (dwChunkSize<=nAvailable)
        goto write;

      break;
    }

    ResetEvent(hEvent);
    YASAPI_PLAYER_UNLOCK_DEEP(pPlayer,&nLockRecursion); // {
    dwWait=WaitForSingleObject(hEvent,msTimeout);
    YASAPI_PLAYER_LOCK_DEEP(pPlayer,nLockRecursion); // }

    switch (dwWait) {
    case WAIT_TIMEOUT:
      switch (FsmState(pClient->pServer)) {
      case eEngineServerStatePauseConnected:
      case eEngineServerStatePauseDisconnected:
        continue;
      default:
        break;
      }

      DVMESSAGE("WAIT_TIMEOUT in state %s",FsmStateId(pClient));
      break;
    case WAIT_FAILED:
      DVMESSAGE("WAIT_FAILED %s: %lu",FsmStateId(pClient),GetLastError());
      goto exit;
    default:
      break;
    }
  }
write:
  code=_EngineClientWrite(pClient,(const BYTE *)buf,dwChunkSize);
exit:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
enoinit:
  return code;
}

int EngineClientApiEndOfTrack(EngineClient *pClient, int msTimeout)
{
  int code=-1;
  Player *pPlayer=pClient->pPlayer;
  HANDLE hEvent=pClient->pServer->api.hEvent;
  int nLockRecursion;
  DWORD dwWait;

  if (FsmState(pClient)<eEngineClientStateBase) {
    DVMESSAGE("unexpected state %s",FsmStateId(pClient));
    goto enoinit;
  }

  YASAPI_PLAYER_LOCK(pPlayer); // {

  for (;;) {
    switch (FsmState(pClient->pServer)) {
    case eEngineServerStateKill:
      goto exit;
    default:
      if (!_EngineClientIsPlaying(pClient)) {
DWRITELN(FsmStateId(pClient));
        goto success;
      }

      break;
    }

    ResetEvent(hEvent);
    YASAPI_PLAYER_UNLOCK_DEEP(pPlayer,&nLockRecursion); // {
    dwWait=WaitForSingleObject(hEvent,msTimeout);
    YASAPI_PLAYER_LOCK_DEEP(pPlayer,nLockRecursion); // }

    switch (dwWait) {
    case WAIT_ABANDONED:
      DVMESSAGE("WAIT_ABANDONED %s",FsmStateId(pClient));
      goto exit;
    case WAIT_TIMEOUT:
      switch (FsmState(pClient->pServer)) {
      case eEngineServerStatePauseConnected:
      case eEngineServerStatePauseDisconnected:
        continue;
      default:
        break;
      }

      DVMESSAGE("WAIT_TIMEOUT %s",FsmStateId(pClient));
      goto exit;
    case WAIT_FAILED:
      DVMESSAGE("WAIT_FAILED %s: %lu",FsmStateId(pClient),GetLastError());
      goto exit;
    default:
      break;
    }
  }
success:
  code=0;
exit:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
enoinit:
  return code;
}

static void _EngineClientApiKill(EngineClient *pClient)
{
  EEngineServer eEngineServer=EngineClientServerState(pClient);

if (eEngineServerStateOpen<=eEngineServer&&eEngineServer<eEngineServerStateDraining)
    YASAPI_REQUEST_STATE(pClient,eEngineServerStateKill);
}

void EngineClientApiKill(EngineClient *pClient)
{
  Player *pPlayer=pClient->pPlayer;

  YASAPI_PLAYER_LOCK(pPlayer); // {
  _EngineClientApiKill(pClient);
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
}

static int EngineClientPreOpen(EngineClient *pClient, int bIncBroken,
    int bIncTracks)
{
  Player *pPlayer=pClient->pPlayer;
  EngineParam *pParam=&pClient->param;
  EngineOptions *pOptions=&pClient->options;

  if (FsmState(pClient)<eEngineClientStateBase) {
    DVMESSAGE("unexpected state %s",FsmStateId(pClient));
    goto exit;
  }

  YASAPI_FSM_DOWN(pClient,eEngineClientStateBase);

  if (EngineParamCreate(pParam,pOptions,pClient)<0) {
    DMESSAGE("creating parameter");
    goto exit;
  }

  // During "open" we thought this is "gapless". Meanwhile the sesssion was
  // terminated.
  if (YASAPI_FSM_UP(pClient,eEngineClientStatePreOpen)<0) {
    DMESSAGE("creating gapless sequence");
    goto exit;
  }

  if (bIncBroken)
    YASAPI_INC_UINT32(pPlayer->stats.uNumGaplessBroken);

  if (bIncTracks) {
    YASAPI_INC_UINT32(pPlayer->stats.uNumTracks);
    YASAPI_INC_UINT32(pClient->stats.uNumTracks);
  }

  if (UINT32_MAX==pPlayer->stats.uNumGaplessBroken)
    pPlayer->stats.uNumGaplessBroken=0lu;

  pClient->stats.uNumFramesDropout=UINT32_MAX;
  YASAPI_INC_UINT32(pPlayer->stats.uNumGapless);
  GtkClientUpdateDialog(&pPlayer->gtk,&pPlayer->gtk.config);

  return 0;
exit:
  return -1;
}

static EEngineServer EngineClientServerState(EngineClient *pClient)
{
  if (FsmState(pClient)<eEngineClientStateBase)
    return eEngineServerStateStart;
  else
    return FsmState(pClient->pServer);
}
