/*
 * yasapi_engine_server_state.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <avrt.h>

///////////////////////////////////////////////////////////////////////////////
#if ! defined (ERROR_INVALID_TASK_NAME) // [
# define ERROR_INVALID_TASK_NAME 1550
#endif // ]
#if ! defined (ERROR_INVALID_TASK_INDEX) // [
# define ERROR_INVALID_TASK_INDEX 1551
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#define PRIO_ALTER(hThread,pbAltered,pnPrevious,ePriority,nRelation,label) \
  EngineServerPriorityAlter(hThread,pbAltered,pnPrevious,ePriority,nRelation)
#define PRIO_RESTORE(hThread,pbAltered,pnPrevious,label) \
  EngineServerPriorityRestore(hThread,pbAltered,pnPrevious)

static int EngineServerPriorityAlter(HANDLE hThread, int *pbAltered,
    int *pnPrevious, EThreadPriority ePriority, int nRelation )
{
  int nPrevious;
  int nPriority;

  *pbAltered=0;

  switch (ePriority) {
  case eThreadPriorityBackground:
    nPriority=THREAD_MODE_BACKGROUND_BEGIN;
    nPrevious=THREAD_MODE_BACKGROUND_END;
    goto alter;
  case eThreadPriorityAboveNormal:
    nPriority=THREAD_PRIORITY_ABOVE_NORMAL;
    break;
  case eThreadPriorityBelowNormal:
    nPriority=THREAD_PRIORITY_BELOW_NORMAL;
    break;
  case eThreadPriorityHighest:
    nPriority=THREAD_PRIORITY_HIGHEST;
    break;
  case eThreadPriorityIdle:
    nPriority=THREAD_PRIORITY_IDLE;
    break;
  case eThreadPriorityLowest:
    nPriority=THREAD_PRIORITY_LOWEST;
    break;
  case eThreadPriorityNormal:
    nPriority=THREAD_PRIORITY_NORMAL;
    break;
  case eThreadPriorityTimeCritical:
    nPriority=THREAD_PRIORITY_TIME_CRITICAL;
    break;
  default:
    DVMESSAGE("unexpected priority: %d",ePriority);
    goto exit;
  }

  nPrevious=GetThreadPriority(
    hThread       // HANDLE hThread
  );

  switch (nRelation) {
  case '<':
    if (nPrevious<=nPriority)
      goto success;

    break;
  case '>':
    if (nPriority<=nPrevious)
      goto success;

    break;
  default:
    DVMESSAGE("unexpected relation %c",nRelation);
    goto exit;
  }

  *pnPrevious=nPrevious;
alter:
  *pbAltered=SetThreadPriority(
    hThread,      // HANDLE hThread,
    nPriority     // int    nPriority
  );

DVWRITELNEX(file,line,func,"!!! %s.bAltered: %d >>>=============",
    label,*pbAltered);
success:
  return 0;
exit:
  return -1;
}

static void EngineServerPriorityRestore(HANDLE hThread, int *pbAltered,
    int *pnPrevious )
{

  if (*pbAltered) {
    *pbAltered=0;

    SetThreadPriority(
      hThread,    // HANDLE hThread,
      *pnPrevious // int    nPriority
    );
  }
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateThreadEventEnter(EngineServer *pServer)
{
  // This is a MANUAL RESET event!
  if (!(pServer->thread.hEvent=CreateEventW(NULL,TRUE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    goto exit;
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStateThreadEventLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
  CloseHandle(pServer->thread.hEvent);
  pServer->thread.hEvent=NULL;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateApiEventEnter(EngineServer *pServer)
{
  // This is a MANUAL RESET event!
  if (!(pServer->api.hEvent=CreateEventW(NULL,TRUE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    goto exit;
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStateApiEventLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
  CloseHandle(pServer->api.hEvent);
  pServer->api.hEvent=NULL;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateKillEnter(EngineServer *pServer)
{
  ++pServer->fsm.pState;

  return 0;
}

static void EngineServerStateKillLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStatePauseDisconnectedEnter(EngineServer *pServer)
{
  ++pServer->fsm.pState;

  return 0;
}

static void EngineServerStatePauseDisconnectedLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateAudioClientEnter(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  IMMDevice *pDevice=pClient->param.pDevice;
  IAudioClient *pAudioClient;
  HRESULT hr;

  if (pClient->pAudioClient) {
    pServer->pAudioClient=pClient->pAudioClient;
    pClient->pAudioClient=NULL;
    ++pServer->fsm.pState;
  }
  else {
    ///////////////////////////////////////////////////////////////////////////
    hr=pDevice->lpVtbl->Activate(pDevice,
      &IID_IAudioClient,      // [in]  REFIID      iid,
      CLSCTX_ALL,             // [in]  DWORD       dwClsCtx,
      NULL,                   // [in]  PROPVARIANT *pActivationParams,
      (void **)&pAudioClient  // [out] void        **ppInterface
    );

    if (FAILED(hr)) {
#define DCLABEL "IMMDevice::Activate<IAudioClient>" // [
      switch (hr) {
      DCMESSAGE(DCLABEL,E_NOINTERFACE);
      DCMESSAGE(DCLABEL,E_POINTER);
      DCMESSAGE(DCLABEL,E_INVALIDARG);
      DCMESSAGE(DCLABEL,E_OUTOFMEMORY);
      DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
      DDMESSAGE(DCLABEL,hr);
      }
#undef DCLABEL // ]
      goto error;
    }

    pServer->pAudioClient=pAudioClient;
    ++pServer->fsm.pState;

    ///////////////////////////////////////////////////////////////////////////
    if (eStrategyPull==pClient->options.device.eStrategy) {
      hr=pAudioClient->lpVtbl->Initialize(pAudioClient,
        pClient->param.eShareMode,
                              // [in]       AUDCLNT_SHAREMODE ShareMode,
        pClient->param.dwStreamFlags,
                              // [in]       DWORD             StreamFlags,
        pClient->param.hnsDevicePeriod,
                              // [in]       REFERENCE_TIME    hnsBufferDuration,
        pClient->param.hnsDevicePeriod,
                              // [in]       REFERENCE_TIME    hnsPeriodicity,
        &pClient->wfx.target.Format,
                              // [in] const WAVEFORMATEX      *pFormat,
        NULL                  // [in]       LPCGUID           AudioSessionGuid
      );
    }
    else {
      hr=pAudioClient->lpVtbl->Initialize(pAudioClient,
        pClient->param.eShareMode,
                              // [in]       AUDCLNT_SHAREMODE ShareMode,
        pClient->param.dwStreamFlags,
                              // [in]       DWORD             StreamFlags,
        pClient->param.hnsDevicePeriod,
                              // [in]       REFERENCE_TIME    hnsBufferDuration,
        0ull,                 // [in]       REFERENCE_TIME    hnsPeriodicity,
        &pClient->wfx.target.Format,
                              // [in] const WAVEFORMATEX      *pFormat,
        NULL                  // [in]       LPCGUID           AudioSessionGuid
      );
    }

    if (FAILED(hr)) {
#define DCLABEL "IAudioClient::Initialize" // [
      switch (hr) {
      DCMESSAGE(DCLABEL,CO_E_NOTINITIALIZED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_ALREADY_INITIALIZED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_WRONG_ENDPOINT_TYPE);
      DCMESSAGE(DCLABEL,AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_BUFFER_SIZE_ERROR);
      DCMESSAGE(DCLABEL,AUDCLNT_E_CPUUSAGE_EXCEEDED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_IN_USE);
      DCMESSAGE(DCLABEL,AUDCLNT_E_ENDPOINT_CREATE_FAILED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_INVALID_DEVICE_PERIOD);
      DCMESSAGE(DCLABEL,AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL);
      DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
      DCMESSAGE(DCLABEL,E_POINTER);
      DCMESSAGE(DCLABEL,E_INVALIDARG);
      DCMESSAGE(DCLABEL,E_OUTOFMEMORY);
      DDMESSAGE(DCLABEL,hr);
      }
#undef DCLABEL // ]
      goto error;
    }

    ///////////////////////////////////////////////////////////////////////////
    if (eStrategyPull==pClient->options.device.eStrategy) {
      hr=pAudioClient->lpVtbl->SetEventHandle(pAudioClient,
        pClient->timer.hEvent   // [in] HANDLE eventHandle
      );

      if (FAILED(hr)) {
#define DCLABEL "IAudioClient::SetEventHandle" // [
        switch (hr) {
        DCMESSAGE(DCLABEL,E_INVALIDARG);
        DCMESSAGE(DCLABEL,AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED);
        DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_INITIALIZED);
        DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
        DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
        DDMESSAGE(DCLABEL,hr);
        }
#undef DCLABEL // ]
        goto error;
      }
    }
  }

  return 0;
error:
  return -1;
}

static void EngineServerStateAudioClientLeave(EngineServer *pServer)
{
  IAudioClient **ppAudioClient=&pServer->pAudioClient;

  --pServer->fsm.pState;
  (*ppAudioClient)->lpVtbl->Release(*ppAudioClient);
  *ppAudioClient=NULL;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateAudioClockEnter(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;

  DASSERT(pClient,exit);

  if (pClient->param.time.lpVtbl->ClockCreate(pServer)<0) {
    goto exit;
  }

  pClient->param.time.lpVtbl->Set(pServer,0);
  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStateAudioClockLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
  pServer->pClient->param.time.lpVtbl->ClockDestroy(pServer);
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateVolumeEnter(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  GtkClient *pGtk;

  DASSERT(pClient,exit);
  DASSERT(pClient->param.volume.lpVtbl,exit);

  if (pClient->param.volume.lpVtbl->Create(pServer)<0)
    goto exit;

  pGtk=&pClient->pPlayer->gtk;
  GtkClientUpdateWidget(pGtk,&pGtk->config.config.monitor.engine.volume);
  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStateVolumeLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
  pServer->pClient->param.volume.lpVtbl->Destroy(pServer);
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateAudioRenderClientEnter(EngineServer *pServer)
{
  IAudioClient *pAudioClient=pServer->pAudioClient;
  HRESULT hr;

  hr=pAudioClient->lpVtbl->GetService(pAudioClient,
    &IID_IAudioRenderClient,    // [in]  REFIID riid,
    (void **)&pServer->pAudioRenderClient
                                // [out] void   **ppv
  );

  if (FAILED(hr)) {
#define DCLABEL "IAudioClient::GetService<IAudioRenderClient>" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,E_POINTER);
    DCMESSAGE(DCLABEL,E_NOINTERFACE);
    DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_INITIALIZED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_WRONG_ENDPOINT_TYPE);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto error;
  }

  ++pServer->fsm.pState;

  return 0;
error:
  return -1;
}

static void EngineServerStateAudioRenderClientLeave(EngineServer *pServer)
{
  IAudioRenderClient *pAudioRenderClient=pServer->pAudioRenderClient;

  --pServer->fsm.pState;
  pAudioRenderClient->lpVtbl->Release(pAudioRenderClient);
  pServer->pAudioRenderClient=NULL;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateFlushEnter(EngineServer *pServer)
{
  ++pServer->fsm.pState;

  return 0;
}

static void EngineServerStateFlushLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStatePrerollEnter(EngineServer *pServer)
{
  UINT32 uFrames=pServer->pClient->param.uNumFramesDevice;


  if (_EngineServerRender(pServer,uFrames,0)<0) {
    DMESSAGE("prerolling");
    goto exit;
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStatePrerollLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStatePauseConnectedEnter(EngineServer *pServer)
{
  ++pServer->fsm.pState;

  return 0;
}

static void EngineServerStatePauseConnectedLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStatePriorityAlterRenderEnter(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;

  if (PRIO_ALTER(pClient->thread.hThread,
      &pServer->priority.bAltered,
      &pServer->priority.nPrevious,
      pClient->options.device.render.ePriority,
      '>',"priority")<0)
    goto exit;

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStatePriorityAlterRenderLeave(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;

  --pServer->fsm.pState;

  PRIO_RESTORE(pClient->thread.hThread,
      &pServer->priority.bAltered,
      &pServer->priority.nPrevious,
      "priority");
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateThreadCharacteristicsEnter(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  DWORD dwLastError;

  if (pClient->options.device.bProAudio&&!pServer->task.hTask) {
    pServer->task.hTask=AvSetMmThreadCharacteristicsW(L"Pro Audio",
        &pServer->task.dwTaskIndex);

    if (!pServer->task.hTask) {
      dwLastError=GetLastError();

#define DCLABEL "AvSetMmThreadCharacteristics" // [
      switch (dwLastError) {
      DCMESSAGE(DCLABEL,ERROR_INVALID_TASK_INDEX);
      DCMESSAGE(DCLABEL,ERROR_INVALID_TASK_NAME);
      DCMESSAGE(DCLABEL,ERROR_PRIVILEGE_NOT_HELD);
      DCMESSAGE(DCLABEL,ERROR_ACCESS_DENIED);
      default:
        DVMESSAGE(DCLABEL " failed with unknow error %lu",dwLastError);
        break;
      }
#undef DCLABEL // ]
      goto error;
    }
  }

  ++pServer->fsm.pState;

  return 0;
error:
  return -1;
}

static void EngineServerStateThreadCharacteristicsLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;

  if (pServer->task.hTask) {
    AvRevertMmThreadCharacteristics(pServer->task.hTask);
    memset(&pServer->task,0,sizeof pServer->task);
  }
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateAudioClientStartEnter(EngineServer *pServer)
{
  IAudioClient *pAudioClient=pServer->pAudioClient;
  HRESULT hr;

  hr=pAudioClient->lpVtbl->Start(pAudioClient);

  if (FAILED(hr)) {
#define DCLABEL "IAudioClient::Start" // [
    switch (hr) {
      DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_INITIALIZED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_STOPPED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_EVENTHANDLE_NOT_SET);
      DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
      DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto error;
  }

  ++pServer->fsm.pState;

  return 0;
error:
  return -1;
}

static void EngineServerStateAudioClientStartLeave(EngineServer *pServer)
{
  IAudioClient *pAudioClient=pServer->pAudioClient;

  --pServer->fsm.pState;
  pAudioClient->lpVtbl->Stop(pAudioClient);
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateSetWaitableTimerEnter(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  BOOL bCode;

  if (eStrategyPush==pClient->options.device.eStrategy) {
    bCode=SetWaitableTimer(
      pClient->timer.hTimer,
                  // _In_           HANDLE           hTimer,
      &pClient->param.DueTime,
                  // _In_     const LARGE_INTEGER    *pDueTime,
      pClient->param.lPeriod,
                  // _In_           LONG             lPeriod,
      NULL,       // _In_opt_       PTIMERAPCROUTINE pfnCompletionRoutine,
      NULL,       // _In_opt_       LPVOID           lpArgToCompletionRoutine,
      FALSE       // _In_           BOOL             fResume
    );

    if (!bCode) {
      DVMESSAGE("starting waitable timer: 0x%04lx",GetLastError());
      goto exit;
    }
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStateSetWaitableTimerLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
    CancelWaitableTimer(pServer->pClient->timer.hTimer);
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateRenderEnter(EngineServer *pServer)
{
  if (EngineServerRender(pServer,FsmState(pServer),eEngineServerStateStart)<0) {
    DMESSAGE("rendering");
    goto exit;
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStateRenderLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStatePriorityAlterWinampEnter(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  HANDLE hThread=pClient->winamp.hThread;

  if (pClient->options.common.bGapless) {
    if (PRIO_ALTER(hThread,
        &pClient->winamp.bAltered,
        &pClient->winamp.nPrevious,
        pClient->options.common.winamp.ePriority,
        '<',"winamp")<0)
      goto exit;
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStatePriorityAlterWinampLeave(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  HANDLE hThread=pClient->winamp.hThread;

  --pServer->fsm.pState;

  PRIO_RESTORE(hThread,
      &pClient->winamp.bAltered,
      &pClient->winamp.nPrevious,
      "winamp");

}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStatePriorityAlterDrainEnter(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  HANDLE hThread=pClient->thread.hThread;

  if (pClient->options.common.bGapless) {
    if (PRIO_ALTER(hThread,
        &pClient->drain.bAltered,
        &pClient->drain.nPrevious,
        pClient->options.common.drain.ePriority,
        '>',"drain")<0)
      goto exit;

/*
    SetThreadPriorityBoost(
      hThread,      // HANDLE hThread,
      TRUE          // BOOL   bDisablePriorityBoost
    );
*/
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStatePriorityAlterDrainLeave(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  HANDLE hThread=pClient->thread.hThread;

  --pServer->fsm.pState;

  if (pClient->drain.bAltered) {
/*
    SetThreadPriorityBoost(
      hThread,      // HANDLE hThread,
      FALSE         // BOOL   bDisablePriorityBoost
    );
*/

    PRIO_RESTORE(hThread,
        &pClient->drain.bAltered,
        &pClient->drain.nPrevious,
        "drain");
  }
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateDrainingEnter(EngineServer *pServer)
{
  ++pServer->fsm.pState;

  return 0;
}

static void EngineServerStateDrainingLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateEotEnter(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  BOOL bCode;

  if (eStrategyPull==pClient->options.device.eStrategy)
    goto success;
  else if (pClient->options.device.bTimerPeriodic)
    goto success;

  bCode=SetWaitableTimer(
    pClient->timer.hTimer,
                // _In_           HANDLE           hTimer,
    &pClient->param.DueTime,
                // _In_     const LARGE_INTEGER    *pDueTime,
    0l,         // _In_           LONG             lPeriod,
    NULL,       // _In_opt_       PTIMERAPCROUTINE pfnCompletionRoutine,
    NULL,       // _In_opt_       LPVOID           lpArgToCompletionRoutine,
    FALSE       // _In_           BOOL             fResume
  );

  if (!bCode) {
    DVMESSAGE("starting waitable timer: 0x%04lx",GetLastError());
    goto exit;
  }
success:
  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineServerStateEotLeave(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  Player *pPlayer=pClient->pPlayer;
  GtkClient *pGtk=&pPlayer->gtk;
  ConfigMonitor *pMonitor=&pGtk->config.config.monitor;

  --pServer->fsm.pState;
  pClient->monitor.xWritten=0.0;
  GtkClientUpdateWidget(pGtk,&pMonitor->written);
  pClient->monitor.xPadding=0.0;
  GtkClientUpdateWidget(pGtk,&pMonitor->padding);
  pClient->monitor.xStart=0.0;
  GtkClientUpdateWidget(pGtk,&pMonitor->start);
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateLoopEndEnter(EngineServer *pServer)
{
  DMESSAGE("illegal attempt entering eEngineServerStateLoopEnd");

  return -1;
}

static void EngineServerStateLoopEndLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int EngineServerStateErrorEnter(EngineServer *pServer)
{
  DMESSAGE("illegal attempt entering eEngineServerStateError");

  return -1;
}

static void EngineServerStateErrorLeave(EngineServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
IEngineServerState gcaIEngineServerState[eEngineServerStateCount]={
#define YASAPI_ENGINE_SERVER(x) { \
  .pId=#x, \
  .Enter=EngineServerState ## x ## Enter, \
  .Leave=EngineServerState ## x ## Leave, \
},
#include <yasapi_engine_server_state.h>
#undef YASAPI_ENGINE_SERVER
};
